import psutil
from tornado import gen, locks, queues, concurrent
from tornado.ioloop import IOLoop

from . import async_process


class BaseTask(object):
    def __init__(self):
        self._canceled = False

    def __next__(self):
        raise StopIteration

    def start(self):
        pass

    def done(self):
        pass

    def cancel(self):
        self._canceled = True


class JobFuture(concurrent.Future):
    def __init__(self, task):
        super(JobFuture, self).__init__()
        self.task = task

    @property
    def id(self):
        return self.task.id

    def cancel(self):
        return self.task.cancel()


class JobQueue(object):
    def __init__(self, threads=None, ioloop=None):
        self._threads = threads if threads is not None else psutil.cpu_count(logical=False)
        self._ioloop = ioloop if ioloop is not None else IOLoop.current()

        self._queue = queues.Queue()
        self._current = queues.Queue()
        self._sem = locks.Semaphore(self._threads)

        self._jobs = 0
        self._finished = locks.Event()
        self._finished.set()

        self._start()

    def _start(self):
        self._ioloop.spawn_callback(self._move_jobs)
        for _ in range(self._threads):
            self._ioloop.spawn_callback(self._worker)

    def _increment_jobs(self):
        self._jobs += 1
        self._finished.clear()

    def _decrement_jobs(self):
        self._jobs -= 1
        if self._jobs <= 0:
            self._finished.set()

    @gen.coroutine
    def _put_wrapper(self, task):
        yield self._queue.put(task)
        self._increment_jobs()
        raise gen.Return(task)

    @gen.coroutine
    def put(self, task):
        fut = JobFuture(task)
        gen.chain_future(self._put_wrapper(task), fut)

        return (yield fut)

    @gen.coroutine
    def _move_jobs(self):
        while True:
            yield self._sem.acquire()
            task = yield self._queue.get()
            if task.start():
                yield self._current.put(task)

    @gen.coroutine
    def _worker(self):
        while True:
            yield self._one_shot_worker()

    def _release(self, task):
        self._sem.release()
        self._decrement_jobs()
        task.done()
        raise gen.Return()

    @gen.coroutine
    def _wrap_dispatcher(self, fut, dispatcher):
        try:
            return (yield dispatcher())
        finally:
            self._decrement_jobs()
            fut.set_result(None)

    @gen.coroutine
    def _spawn_dispatcher(self, dispatcher):
        done = concurrent.Future()
        self._ioloop.spawn_callback(self._wrap_dispatcher, done, dispatcher)
        yield done

    @gen.coroutine
    def _one_shot_worker(self):
        task = yield self._current.get()
        if task._canceled:
            self._release()

        try:
            dispatcher = next(task)
            self._increment_jobs()

        except StopIteration:
            self._release(task)

        yield self._current.put(task)
        yield self._spawn_dispatcher(dispatcher)

    @gen.coroutine
    def join(self):
        yield self._finished.wait()


@gen.coroutine
def _main():
    import sys

    class MultithreadTask(BaseTask):
        def __init__(self, args):
            super(MultithreadTask, self).__init__()
            self._iter = iter(args)

        def __next__(self):
            v = next(self._iter)

            @gen.coroutine
            def dispatcher():
                def in_subprocess(v):
                    if v == 4:
                        raise ValueError(v)
                    import time
                    time.sleep(1)
                    return v

                r = yield async_process.dispatch(in_subprocess, (v,))
                print(r)

            return dispatcher

    threads = None if len(sys.argv) <= 1 else int(sys.argv[1])

    q = JobQueue(threads=threads)
    Task = MultithreadTask
    yield q.put(Task([1, 2, 3, 4, 5]))
    yield q.put(Task([6, 7, 8]))
    yield q.put(Task([9, 10, 11, 12, 13, 14]))

    yield q.join()


if __name__ == '__main__':
    IOLoop.current().run_sync(_main)



import os

import tornado.web

from .db import connect
from .job_queue import JobQueue
from .handler.calc import CalcIdHandler, CalcIdExtHandler
from .handler.file import FileHandler, FileIdHandler, FileIdExtHandler, FileIdNthExtHandler
from .handler.descriptor import DescriptorHandler
from .handler.singlefile import SingleFileHandler


class MyApplication(tornado.web.Application):
    def __init__(self, queue, conn, *args, **kwargs):
        super(MyApplication, self).__init__(*args, **kwargs)
        self.queue = queue
        self.db = conn


def main(port, workers, db='mordred-web.sqlite'):
    static = os.path.join(os.path.dirname(__file__), 'static')

    with connect(db) as conn, JobQueue(workers) as queue:
        app = MyApplication(queue, conn, [
            (r'/api/descriptor', DescriptorHandler),
            (r'/api/file', FileHandler),
            (r'/api/file/([0-9a-zA-Z]+)', FileIdHandler),
            (r'/api/file/([0-9a-zA-Z]+)\.(.*)', FileIdExtHandler),
            (r'/api/file/([0-9a-zA-Z]+)/([0-9]+)\.(.*)', FileIdNthExtHandler),
            (r'/api/calc/([0-9a-zA-Z]+)', CalcIdHandler),
            (r'/api/calc/([0-9a-zA-Z]+)\.(.*)', CalcIdExtHandler),
            (r'/static/(.*)', tornado.web.StaticFileHandler, {'path': static}),
            (r'/.*', SingleFileHandler, {'path': os.path.join(static, 'index.html')}),
        ], compress_response=True, static_hash_cache=True)
        server = tornado.httpserver.HTTPServer(app)
        server.bind(port)
        print('start mordred.web on localhost:{}'.format(port))
        server.start(1)
        tornado.ioloop.IOLoop.instance().start()


if __name__ == "__main__":
    main(3000, 1)



from __future__ import print_function

import math
import pickle
import multiprocessing
from abc import ABCMeta, abstractmethod
from collections import Counter

from six import with_metaclass
from tornado import gen, queues
from tornado.ioloop import IOLoop


class Worker(with_metaclass(ABCMeta, object)):
    @abstractmethod
    def __next__(self):
        raise NotImplementedError

    @abstractmethod
    def __call__(self):
        raise NotImplementedError


class SingleWorker(Worker):
    def __next__(self):
        if getattr(self, '_done', False) is True:
            raise StopIteration

        self._done = True


class Task(with_metaclass(ABCMeta, object)):
    def on_start(self):
        pass

    def on_end(self):
        pass

    def on_task_start(self):
        pass

    def on_task_end(self, result):
        pass

    def on_task_error(self, err):
        pass

    def next_task(self):
        return None

    @abstractmethod
    def worker(self):
        raise NotImplementedError


class JobQueue(object):
    START = 0
    END = 1
    TASK_START = 2
    TASK_END = 3
    TASK_ERROR = 4

    _current_task_id = 0

    def __init__(self, threads, ioloop=None):
        if ioloop is None:
            ioloop = IOLoop.current()

        self._ioloop = ioloop

        self._threads = threads
        self._pending = multiprocessing.Queue()
        self._working = multiprocessing.Queue(threads)
        self._semaphore = multiprocessing.Semaphore(threads)
        self._task_count = 0
        self._task_done = queues.Queue()
        self._events_out, self._events_in = multiprocessing.Pipe(False)
        self._task_counter = Counter()
        self._tasks = {}

    def put(self, task):
        self._task_count += 1
        task_id = self.__class__._current_task_id
        self.__class__._current_task_id += 1
        self._tasks[task_id] = task
        self._pending.put((task.worker(), task_id))

    def _move_worker(self):
        while True:
            self._semaphore.acquire()
            worker, task_id = self._pending.get()
            self._events_in.send((self.START, task_id, None))
            self._working.put((worker, task_id))

    def _exec_worker(self, worker):
        result_out, result_in = multiprocessing.Pipe(False)
        chunk_size = 10240

        def f():
            try:
                v = worker()
            except Exception as e:
                data = True, e
            else:
                data = False, v

            b = pickle.dumps(data)

            for i in range(math.ceil(len(b) / chunk_size)):
                st = i * chunk_size
                ed = st + chunk_size
                result_in.send_bytes(b[st:ed])

            result_in.send_bytes(b'')

        p = multiprocessing.Process(target=f)
        p.start()

        msg = b''
        while True:
            b = result_out.recv_bytes()
            if len(b) == 0:
                break

            msg += b

        p.join()
        return pickle.loads(msg)

    def _job_worker(self):
        while True:
            worker, task_id = self._working.get()
            try:
                worker.__next__()
            except StopIteration:
                self._events_in.send((self.END, task_id, None))
                continue
            else:
                self._working.put((worker, task_id))

            self._events_in.send((self.TASK_START, task_id, None))

            err, v = self._exec_worker(worker)
            self._events_in.send((self.TASK_ERROR if err else self.TASK_END, task_id, v))

    @gen.coroutine
    def _event_watcher(self, fd, events):
        ev, task_id, val = fd.recv()
        task = self._tasks[task_id]

        if ev == self.START:
            task.on_start()
            self._task_counter[task_id] += 1
            return

        if ev == self.TASK_START:
            task.on_task_start()
            self._task_counter[task_id] += 1
            return

        self._task_counter[task_id] -= 1
        if ev == self.TASK_END:
            task.on_task_end(val)

        elif ev == self.TASK_ERROR:
            task.on_task_error(val)

        if self._task_counter[task_id] <= 0:
            task.on_end()
            next_task = task.next_task()
            if next_task is not None:
                self.put(next_task)

            del self._task_counter[task_id]
            del self._tasks[task_id]
            self._semaphore.release()
            yield self._task_done.put(None)

    def __enter__(self):
        self._processes = processes = [multiprocessing.Process(target=self._move_worker)]
        for _ in range(self._threads):
            processes.append(multiprocessing.Process(target=self._job_worker))

        for p in processes:
            p.start()

        self._ioloop.add_handler(self._events_out, self._event_watcher, IOLoop.READ)
        return self

    def __exit__(self, *args, **kwargs):
        self._ioloop.remove_handler(self._events_out)
        for p in self._processes:
            p.terminate()

    @gen.coroutine
    def join(self):
        while True:
            yield self._task_done.get()
            self._task_count -= 1
            if self._task_count <= 0:
                return


class TestTask(Task):
    def __init__(self, name, u, i=4):
        self.name = name
        self.u = u
        self.i = i

    def __call__(self):
        import time
        import random
        v = random.uniform(0, self.u)
        time.sleep(v)
        if self.name == "ERROR":
            raise ValueError("VALUE ERROR")

        return v

    def on_task_start(self):
        print('TASK START', self.name)

    def on_task_end(self, v):
        print('TASK END', self.name, v)

    def on_start(self):
        print("START", self.name)

    def on_end(self):
        print("END", self.name)

    def on_task_error(self, e):
        print("TASK ERROR", self.name, e)

    def __next__(self):
        if self.i <= 0:
            raise StopIteration

        self.i -= 1

    def worker(self):
        return self


@gen.coroutine
def main():
    import sys
    _, t, n, u = sys.argv
    with JobQueue(int(t)) as queue:
        for i in range(int(n)):
            queue.put(TestTask(i, float(u)))
        queue.put(TestTask("ERROR", float(u)))
        yield queue.join()
        print("FINISH")


if __name__ == '__main__':
    IOLoop.current().run_sync(main)



from tornado import gen, web, iostream
from contextlib import contextmanager
from ..db import transaction
import json


class RequestHandler(web.RequestHandler):
    @contextmanager
    def transaction(self):
        with transaction(self.db) as cur:
            yield cur

    def get_flag(self, name, default=object()):
        v = self.get_argument(name, default)

        if isinstance(v, bool):
            return v

        if v.lower() in ["false", "f"]:
            return False

        return True

    def put(self, task):
        return self.application.queue.put(task)

    @property
    def db(self):
        return self.application.db

    def fail(self, status, reason):
        raise web.HTTPError(status, reason=reason)

    def json(self, **kwargs):
        return self.write(kwargs)


class SSEHandler(RequestHandler):
    def init_sse(self):
        self.set_header('content-type', 'text/event-stream')
        self.set_header('cache-control', 'no-cache')

    @gen.coroutine
    def publish(self, **obj):
        self.write('data: {}\n\n'.format(json.dumps(obj)))
        try:
            yield self.flush()
        except iostream.StreamClosedError:
            raise web.Finish



import time
from cgi import parse_header
import math

from mordred import Calculator, descriptors
from tornado import gen, web
from mordred.error import MissingValueBase

from ..db import transaction, issue_text_id
from .common import RequestHandler, SSEHandler
from ..job_queue import Task, Worker, SingleWorker


class PrepareTask(Task):
    def __init__(self, calc_id, total, file_id, disabled, conn):
        self.calc_id = calc_id
        self.file_id = file_id
        self.disabled = disabled
        self.conn = conn
        self.total = total

    def worker(self):
        return PrepareWorker(disabled=self.disabled)

    def on_task_error(e):
        print(e)

    def on_task_end(self, calc):
        self.calc = calc
        self.desc_ids = []

        with transaction(self.conn) as cur:
            for desc in calc.descriptors:
                cur.execute(
                    'INSERT INTO descriptor (calc_id, name) VALUES (?, ?)',
                    (self.calc_id, str(desc))
                )
                self.desc_ids.append(cur.lastrowid)

    def next_task(self):
        task = CalcTask(
            file_id=self.file_id, calc_id=self.calc_id, desc_ids=self.desc_ids,
            total=self.total, calc=self.calc, conn=self.conn
        )
        task.get_mols()
        return task


class PrepareWorker(SingleWorker):
    def __init__(self, disabled):
        self.disabled = disabled

    def __call__(self):
        calc = Calculator(
            getattr(descriptors, d)
            for d in descriptors.__all__
            if d not in self.disabled
        )

        return calc


class CalcTask(Task):
    def __init__(self, file_id, calc_id, desc_ids, calc, total, conn):
        self.file_id = file_id
        self.calc_id = calc_id
        self.desc_ids = desc_ids
        self.calc = calc
        self.conn = conn
        self.total = total

        Nd = len(desc_ids)
        self.max = [None] * Nd
        self.min = [None] * Nd
        self.mean = [None] * Nd
        self.M = [0.0] * Nd
        self.S = [0.0] * Nd
        self.k = [0] * Nd

    def get_mols(self):
        with transaction(self.conn) as cur:
            cur.execute(
                'SELECT id, mol FROM molecule WHERE file_id = ?',
                (self.file_id,)
            )
            self.mols = cur.fetchall()

    def worker(self):
        return CalcWorker(mols=self.mols, calc=self.calc)

    def on_task_error(self, e):
        print(e)

    def on_end(self):
        std = ((None if k == 0 else math.sqrt(S / k)) for S, k in zip(self.S, self.k))
        results = zip(self.desc_ids, self.min, self.max, self.mean, std)

        with transaction(self.conn) as cur:
            for desc_id, vmin, vmax, mean, std in results:
                cur.execute(
                    'UPDATE descriptor SET min = ?, max = ?, mean = ?, std = ? WHERE id = ?',
                    (vmin, vmax, mean, std, desc_id)
                )

            cur.execute('UPDATE calc SET done = 1 WHERE id = ?', (self.calc_id,))

    def on_task_end(self, results):
        mol_id, results = results

        with transaction(self.conn) as cur:
            for i, (desc_id, result) in enumerate(zip(self.desc_ids, results)):
                value, error = None, None
                if isinstance(result, MissingValueBase):
                    error = str(result.error)
                else:
                    value = result

                cur.execute('''
                    INSERT INTO result (calc_id, molecule_id, descriptor_id, value, error)
                    VALUES (?, ?, ?, ?, ?)
                    ''', (self.calc_id, mol_id, desc_id, value, error))

                if error:
                    continue

                if self.max[i] is None or self.max[i] < value:
                    self.max[i] = value

                if self.min[i] is None or self.min[i] > value:
                    self.min[i] = value

                self.mean[i] = (self.mean[i] or 0) + value / self.total

                self.k[i] += 1

                M = self.M[i]
                self.M[i] += (value - M) / self.k[i]
                self.S[i] += (value - M) * (value - self.M[i])

            cur.execute('UPDATE calc SET current = current + 1 WHERE id = ?', (self.calc_id,))


class CalcWorker(Worker):
    def __init__(self, mols, calc):
        self.mols = mols
        self.calc = calc

    def __next__(self):
        if len(self.mols) == 0:
            raise StopIteration

        self.mol, self.mols = self.mols[0], self.mols[1:]

    def __call__(self):
        mol_id, mol = self.mol
        result = self.calc(mol)
        return mol_id, result


class CalcIdHandler(SSEHandler):
    def post(self, file_text_id):
        with self.transaction() as cur:
            cur.execute(
                'SELECT id, name, total FROM file WHERE text_id = ? LIMIT 1',
                (file_text_id,),
            )
            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            file_id, file_name, total = result

            calc_text_id = issue_text_id()

            cur.execute('''
            INSERT INTO calc (file_id, text_id, created_at, current, done)
            VALUES (?, ?, ?, 0, 0)''', (file_id, calc_text_id, int(time.time())))

            calc_id = cur.lastrowid

        disabled = set(self.get_arguments("disabled"))
        task = PrepareTask(
            calc_id=calc_id, file_id=file_id, total=total, disabled=disabled, conn=self.db
        )
        self.put(task)

        self.json(id=calc_text_id)

    def get(self, calc_text_id):
        with self.transaction() as cur:
            cur.execute('SELECT id, file_id FROM calc WHERE text_id = ? LIMIT 1', (calc_text_id,))
            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            self.calc_id, self.file_id = result

            cur.execute(
                'SELECT name, total, text_id FROM file WHERE id = ? LIMIT 1',
                (self.file_id,),
            )
            self.file_name, self.total, file_text_id = cur.fetchone()

        accept, _ = parse_header(self.request.headers['Accept'])

        if accept == 'text/event-stream':
            return self.get_sse()
        else:
            return self.get_json()

    @gen.coroutine
    def get_sse(self):
        self.init_sse()
        while True:
            with self.transaction() as cur:
                cur.execute(
                    'SELECT done, current FROM calc WHERE id = ? LIMIT 1',
                    (self.calc_id,),
                )

                done, current = cur.fetchone()

            yield self.publish(
                total=self.total, name=self.file_name, done=bool(done), current=current
            )

            if done:
                raise web.Finish
            yield gen.sleep(0.5)

    def get_json(self):
        with self.transaction() as cur:
            cur.execute(
                'SELECT name, text_id FROM file WHERE id = ? LIMIT 1',
                (self.file_id,),
            )

            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            cur.execute('''
                SELECT name, max, min, mean, std
                FROM descriptor
                WHERE calc_id = ?
                ORDER BY id
            ''', (self.calc_id,))

            descs = [
                {"name": name, "max": vmax, "min": vmin, "mean": mean, "std": std}
                for name, vmax, vmin, mean, std in cur.fetchall()
            ]

            file_name, file_text_id = result

            self.json(
                file_name=file_name,
                file_id=file_text_id,
                descriptors=descs,
            )


class CalcIdExtHandler(RequestHandler):
    EXTS = set(["csv"])

    def get(self, calc_text_id, ext):
        if ext not in self.EXTS:
            self.fail(400, "unknown extension")

        with self.transaction() as cur:
            cur.execute('SELECT id, file_id FROM calc WHERE text_id = ? LIMIT 1', (calc_text_id,))
            result = cur.fetchone()
            if result is None:
                self.fail(404, "no id")

            self.calc_id, self.file_id = result

            cur.execute(
                'SELECT id, name FROM molecule WHERE file_id = ? ORDER BY nth',
                (self.file_id,)
            )
            self.molecules = cur.fetchall()

            cur.execute(
                'SELECT name FROM descriptor WHERE calc_id = ? ORDER BY id',
                (self.calc_id,)
            )
            self.descriptors = [d for d, in cur.fetchall()]

            if ext == "csv":
                return self.get_csv(cur)

    def get_csv(self, cur):
        self.set_header('content-type', 'text/csv')
        self.write('name,')
        self.write(','.join(self.descriptors))
        self.write('\n')

        for mol_id, name in self.molecules:
            self.write(name + ',')
            cur.execute('''
                SELECT value FROM result WHERE calc_id = ? AND molecule_id = ?
                ORDER BY descriptor_id
                ''', (self.calc_id, mol_id))

            self.write(','.join(('' if v is None else str(v)) for v, in cur.fetchall()))
            self.write('\n')



import os
from tornado.web import StaticFileHandler


class SingleFileHandler(StaticFileHandler):
    def initialize(self, path):
        super(SingleFileHandler, self).initialize(os.path.dirname(path))
        self.file = os.path.basename(path)

    def head(self):
        return super(SingleFileHandler, self).head(self.file)

    def get(self):
        return super(SingleFileHandler, self).get(self.file)



from .common import RequestHandler
from mordred import descriptors


class DescriptorHandler(RequestHandler):
    def get(self):
        self.write({
            'descriptors': descriptors.__all__
        })
